ubiquitous_classes(['ubiquitous']).

essential_classes(['amino acid','nucleic acid','polysacharide','membrane','intermediate']).

wild_type_knockouts(['YEL021W','YLR303W','YCL018W','YOR202W']).

reverse_operator(OP) :-
	OP == '<->'.

reverse_operator(OP) :-
	OP == '<=>'.


exp_generator_directory('/home/knw/Robot_Scientist/KEGG/ligand/prolog_ligand/experiment_generator/').


%%%%%%%%%%%%%%%%%%  forwards
can_run_reaction(R,NotAllowed,Cell,NewCell,Direction) :-
	R = reaction(RNum,LHS,_OP,RHS),
	\+member(RNum,NotAllowed),  % runs allowing reactions - 
				% remove to run with all reactions allowed
	%\+disallowed_mitochondrion_reaction(RNum),
	check_all_reactants(LHS,Cell),
	add_all_reactants(RHS,Cell,NewCell),
	Direction = forwards.



%%%%%%%%%%%%%%%%  reverse

can_run_reaction(R,NotAllowed,Cell,NewCell,Direction) :-
	R = reaction(RNum,LHS,OP,RHS),
	reverse_operator(OP),
	\+member(RNum,NotAllowed),	   % runs allowing reactions - 
                                   % remove to run with all reactions allowed
        %\+disallowed_mitochondrion_reaction(RNum),
	check_all_reactants(RHS,Cell),
	add_all_reactants(LHS,Cell,NewCell),
	Direction = reverse.


disallowed_mitochondrion_reaction(RNum) :-
	mitochondrion_reaction(RNum),
	reaction_match(RNum,_Other).

run_reactions([],_Allowed,UpdatedCell,UpdatedCell,_A,Ran,Ran).


run_reactions([R|Reactions],Allowed,Cell,UpdatedCell,AlreadyRun,ADone,Ran) :-
	R = reaction(RNum,LHS,OP,RHS),
	\+member(RNum,AlreadyRun),
	can_run_reaction(R,Allowed,Cell,NewCell,Direction),
	write('running':RNum:LHS:OP:RHS:Direction),nl,
	!,
	run_reactions(Reactions,Allowed,NewCell,UpdatedCell,AlreadyRun,[RNum|ADone],Ran).


run_reactions([_R|Reactions],Allowed,Cell,UpdatedCell,AlreadyRun,ADone,Ran) :-
	!,run_reactions(Reactions,Allowed,Cell,UpdatedCell,AlreadyRun,ADone,Ran).




run_cell(stop,_R,_Allowed,Cell,FinalCell,Ran,Ran) :-
	Cell = cell(External,Cytosol,Mitochondrion),
	make_set(External,[],FinalExternal),
	make_set(Cytosol,[],FinalCytosol),
	make_set(Mitochondrion,[],FinalMitochondrion),
	FinalCell = cell(FinalExternal,FinalCytosol,FinalMitochondrion).



run_cell(continue,Reactions,Allowed,Cell,FinalCell,AlreadyRun,FinalRan) :-
	run_reactions(Reactions,Allowed,Cell,UpdatedCell,AlreadyRun,[],Ran),
	write('finished reaction cycle - ran'),nl,nl,
	write(Ran),nl,%get0(10),	
	conc(Ran,AlreadyRun,NewAlreadyRun),
	determine_action(Ran,Action),!, 
	run_cell(Action,Reactions,Allowed,UpdatedCell,FinalCell,NewAlreadyRun,FinalRan).





determine_action([],stop).

determine_action(_List,continue).


knockouts_for_wild_type([],Reactions,Reactions).

knockouts_for_wild_type([WK|WildKnockouts],Done,Reactions) :-
	find_knocked_out_reactions_and_enzymes(WK,KR,_KE),
	conc(Done,KR,NewDone),!,
	knockouts_for_wild_type(WildKnockouts,NewDone,Reactions).


run_yeast_reactions(Knockout,Added,NotAllowed,NotReached,Ran) :-
	all_model_reactions(Reactions),%trace,
        wild_type_knockouts(WildKnockouts),
	knockouts_for_wild_type(WildKnockouts,[],WKReactions),
	find_knocked_out_reactions_and_enzymes(Knockout,
					    RemovedReactions,_RemovedEnzymes),
	conc(WKReactions,RemovedReactions,AllRemovedReactions),
	conc(AllRemovedReactions,NotAllowed,Avoid),
	write('not allowed':Avoid),nl,get0(10),
	create_cell_start(CellStart,Added),%write('about to run cell'),%nl,nl,trace,
	run_cell(continue,Reactions,Avoid,CellStart,FinalCell,[],Ran),%write('ran cell'),nl,nl,
	FinalCell = cell(_External,Cytosol,_Mitochondrion),
	essential_compounds(Essentials),
	compounds_in_compartment(Essentials,Cytosol,cytosol,[],NotReached).

member_yeast_reactions(Knockout,Added,MemberList,NotAllowed,NotReached,Ran) :-
	all_model_reactions(Reactions),%trace,
        wild_type_knockouts(WildKnockouts),
	knockouts_for_wild_type(WildKnockouts,[],WKReactions),
	find_knocked_out_reactions_and_enzymes(Knockout,
					    RemovedReactions,_RemovedEnzymes),
	conc(WKReactions,RemovedReactions,AllRemovedReactions),
	conc(AllRemovedReactions,NotAllowed,Avoid),
	create_cell_start(CellStart,Added),%write('about to run cell'),%nl,nl,trace,
	run_cell(continue,Reactions,Avoid,CellStart,FinalCell,[],Ran),%write('ran cell'),nl,nl,
	FinalCell = cell(_External,Cytosol,_Mitochondrion),
	essential_compounds(Essentials),
	compounds_in_compartment(Essentials,Cytosol,cytosol,[],NotReached),
	test_memberships(MemberList,FinalCell).


gene_list('YJR148W',[compartment_member(mitochondrion,'C00022'),
		     compartment_member(mitochondrion,'C00109'),
		     compartment_member(mitochondrion,'ABUT'),
		     compartment_member(mitochondrion,'C04272'),
		     compartment_member(mitochondrion,'C00141'),
		     compartment_member(mitochondrion,'C00123'),
		     compartment_member(mitochondrion,'C00407'),
		     compartment_member(mitochondrion,'C00183')]).


test_membership(CompMember,FinalCell) :-
	CompMember = compartment_member(Compartment,CID),
	Compartment = external,
	FinalCell = cell(External,_C,_M),
	member(CID,External).

test_membership(CompMember,FinalCell) :-
	CompMember = compartment_member(Compartment,CID),
	Compartment = cytosol,
	FinalCell = cell(_E,Cytosol,_M),
	member(CID,Cytosol).

test_membership(CompMember,FinalCell) :-
	CompMember = compartment_member(Compartment,CID),
	Compartment = mitochondrion,
	FinalCell = cell(_E,_C,Mitochondrion),
	member(CID,Mitochondrion).

test_memberships([],_F).


test_memberships([C|CompMembers],FinalCell) :-
	test_membership(C,FinalCell),
	C = compartment_member(Compartment,CID),
	(   l_compound_name(CID,[N|_Names]),
	    write(CID:N),write(' is found in '),write(Compartment)
	;   
	    write(CID),write(' is found in '),write(Compartment)),
	nl,!,
	test_memberships(CompMembers,FinalCell).


test_memberships([C|CompMembers],FinalCell) :-
	C = compartment_member(Compartment,CID),
	(   l_compound_name(CID,[N|_Names]),
	    write(CID:N),write(' is not found in '),write(Compartment)
	;   
	    write(CID),write(' is not found in '),write(Compartment)),
	nl,!,
	test_memberships(CompMembers,FinalCell).
	




wild_type_reactions(Reactions) :-
	run_yeast_reactions(none,[],[],_NF,Reactions).


	
%% check external location

reactant_in_cell(Reactant,Cell) :-
	Reactant = reactant(external,_S,Compound),
	Cell = cell(External,_,_),
	member(Compound,External).

%% check Cytosol location

reactant_in_cell(Reactant,Cell) :-
	Reactant = reactant(cytosol,_S,Compound),
	Cell = cell(_,Cytosol,_),
	member(Compound,Cytosol).

%% check Mitochondrion location - changed to cytosol

reactant_in_cell(Reactant,Cell) :-
	Reactant = reactant(mitochondrion,_S,Compound),
	Cell = cell(_,_,Mitochondrion),
	member(Compound,Mitochondrion).
%	Cell = cell(_,Cytosol,_),
%	member(Compound,Cytosol).


%% check external location

reactant_in_cell(Reactant,Cell) :-
	Reactant = reactant(external,_S,Compound),
	more_compound(Compound,MoreCompound),
	Cell = cell(External,_,_),
	member(MoreCompound,External).

%% check Cytosol location

reactant_in_cell(Reactant,Cell) :-
	Reactant = reactant(cytosol,_S,Compound),
	more_compound(Compound,MoreCompound),
	Cell = cell(_,Cytosol,_),
	member(MoreCompound,Cytosol).

%% check Mitochondrion location

reactant_in_cell(Reactant,Cell) :-
	Reactant = reactant(mitochondrion,_S,Compound),
	more_compound(Compound,MoreCompound),
	Cell = cell(_,_,Mitochondrion),
	member(MoreCompound,Mitochondrion).


check_all_reactants([],_C).

check_all_reactants([R|Reactants],Cell) :-
	reactant_in_cell(R,Cell),!,
	check_all_reactants(Reactants,Cell).



%% add to external location

add_reactant_to_cell(Reactant,Cell,NewCell) :-
	Reactant = reactant(Location,_S,Compound),
	Location == 'external',
	Cell = cell(External,Cytosol,Mitochondrion),
	conc([Compound],External,NewExternal),
	NewCell = cell(NewExternal,Cytosol,Mitochondrion).


%% add Cytosol location

add_reactant_to_cell(Reactant,Cell,NewCell) :-
	Reactant = reactant(Location,_S,Compound),
	Location == 'cytosol',
	Cell = cell(External,Cytosol,Mitochondrion),
	conc([Compound],Cytosol,NewCytosol),
	NewCell = cell(External,NewCytosol,Mitochondrion).



%% add Mitochondrion location - changed to cytosol

add_reactant_to_cell(Reactant,Cell,NewCell) :-
	Reactant = reactant(Location,_S,Compound),
	Location == 'mitochondrion',
	Cell = cell(External,Cytosol,Mitochondrion),
	conc([Compound],Mitochondrion,NewMitochondrion),
	NewCell = cell(External,Cytosol,NewMitochondrion).
%	conc([Compound],Cytosol,NewCytosol),
%	NewCell = cell(External,NewCytosol,Mitochondrion).


add_all_reactants([],NewCell,NewCell).

add_all_reactants([R|Reactants],Cell,NewCell) :-
	add_reactant_to_cell(R,Cell,Added),!,
	add_all_reactants(Reactants,Added,NewCell).



findall_starts(Starts) :-
	findall(C,start_compound(_N,C),Starts).

findall_importants(Importants,Class) :-
	findall(C,important_compound(C,_N,Class),Importants).

important_classes([],Importants,Importants).

important_classes([Class|Classes],Done,Importants) :-
	findall_importants(These,Class),
	conc(Done,These,NewDone),
	important_classes(Classes,NewDone,Importants).


ubiquitous_compounds(Ubiquitous) :-
	ubiquitous_classes(UClasses),
	important_classes(UClasses,[],Ubiquitous).


essential_compounds(Essentials) :-
	essential_classes(EClasses),
	important_classes(EClasses,[],Essentials).




all_model_reactions(Reactions) :-
	clauses_from_file('dk_reactions_as_kegg.pl',DKReactions),
	clauses_from_file('added_kegg_reactions.pl',KReactions),
	clauses_from_file('putative_cholesterol_reactions.pl',CReactions),
	clauses_from_file('add_palmitate_reaction.pl',PReactions),
	conc(DKReactions,KReactions,TempReactions),
	conc(TempReactions,PReactions,Temp2Reactions),
	conc(Temp2Reactions,CReactions,Reactions).



display_items(_Max,[]).

display_items(Max,Items) :-
	write_items(Items,Remaining,1,Max),nl,!,
	display_items(Max,Remaining).


write_items([],[],_I,_Max_).

write_items(Items,Items,I,Max) :-
	I > Max.

write_items([Item|Items],Remaining,I,Max) :-
	write(Item),write(' '),
	I2 is I + 1,!,
	write_items(Items,Remaining,I2,Max).


enzymes_and_reactions_for_knockout(Knockout,Enzymes,Reactions) :-
	findall(R,orf_fact(Knockout,_EnzClass,_EnzymeClass,_G,_GD,R),Rs),
	findall(EnzClass,orf_fact(Knockout,EnzClass,_EnzymeClass,_G,_GD,R),Es),
	make_set(Rs,[],Reactions),
	make_set(Es,[],Enzymes).

coded_for_by_other_orf(R,Knockout) :-
	orf_fact(Orf2,_E,_EClass,_G,_GD,R),
	Orf2 \== Knockout.


knocked_out_reactions(_Knockout,[],Reactions,Reactions).

knocked_out_reactions(Knockout,[R|Rs],Found,Reactions) :-
	\+coded_for_by_other_orf(R,Knockout),!,
	knocked_out_reactions(Knockout,Rs,[R|Found],Reactions).

knocked_out_reactions(Knockout,[_R|Rs],Found,Reactions) :-
        !,knocked_out_reactions(Knockout,Rs,Found,Reactions).


find_knocked_out_reactions_and_enzymes(none,[],[]).


find_knocked_out_reactions_and_enzymes(Knockout,KnockedOutReactions,KnockedOutEnzymes) :-
	enzymes_and_reactions_for_knockout(Knockout,_CodedEnzymes,CodedReactions),!,
	knocked_out_reactions(Knockout,CodedReactions,[],KnockedOutReactions),
	enzyme_set_for_reactions(KnockedOutReactions,KnockedOutEnzymes).



enzyme_set_for_reactions(Rs,EnzymeSet) :-
	enzymes_for_reactions(Rs,[],Enzymes),
	make_set(Enzymes,[],EnzymeSet).


enzymes_for_reactions([],Es,Es).

enzymes_for_reactions([R|Rs],Found,Es) :-
	findall(EnzClass,orf_fact(_Orf,EnzClass,_EClass,_G,_GD,R),Ezs),
	conc(Found,Ezs,NewFound),!,
	enzymes_for_reactions(Rs,NewFound,Es).



all_reactants_with_locations([],R,R).

all_reactants_with_locations([R|Reactants],Done,RLocs) :-
	R = reactant(Loc,_S,CID),
	NewR = reactant_location(Loc,CID),!,
	all_reactants_with_locations(Reactants,[NewR|Done],RLocs).


reactant_locations_from_reactions([],RLs,RLs).

reactant_locations_from_reactions([R|Reactions],Done,RLs) :-
	R = reaction(_RNum,LHS,_OP,RHS),
	conc(LHS,RHS,AllReactants),
	all_reactants_with_locations(AllReactants,[],RLocs),
	conc(Done,RLocs,NewDone),!,
	reactant_locations_from_reactions(Reactions,NewDone,RLs).


populate_cell([],External,External,Cytosol,Cytosol,
               Mitochondrion,Mitochondrion).

populate_cell([R|Reactants],RExternal,External,RCytosol,Cytosol,
               RMitochondrion,Mitochondrion) :-
	R = reactant(Loc,CID),
	Loc == external,!,
	populate_cell(Reactants,[CID|RExternal],External,RCytosol,
		      Cytosol,RMitochondrion,Mitochondrion).

populate_cell([R|Reactants],RExternal,External,RCytosol,Cytosol,
               RMitochondrion,Mitochondrion) :-
	R = reactant(Loc,CID),
	Loc == cytosol,!,
	populate_cell(Reactants,RExternal,External,[CID|RCytosol],
		      Cytosol,RMitochondrion,Mitochondrion).

populate_cell([R|Reactants],RExternal,External,RCytosol,Cytosol,
               RMitochondrion,Mitochondrion) :-
	R = reactant(Loc,CID),
	Loc == mitochondrion,!,
	populate_cell(Reactants,RExternal,External,RCytosol,
		      Cytosol,[CID|RMitochondrion],Mitochondrion).



create_cell_start(CellStart,Added) :-
	findall_starts(Starts),
	conc(Starts,Added,AddedCompounds),
	ubiquitous_compounds(Ubiquitous),
	conc(AddedCompounds,Ubiquitous,StartCompounds),
	CellStart = cell([],StartCompounds,StartCompounds).



more_compound(Compound,MoreCompound) :-
	multiple_atom_concat(['more ',Compound],MoreCompound).

compound_from_more(MoreCompound,Compound) :-
	name(MoreCompound,[M,O,R,E,_Space|ACompound]),
	name('more',[M,O,R,E]),
	name(Compound,ACompound).


create_more_compounds([],More,More).

create_more_compounds([C|Compounds],Done,More) :-
	more_compound(C,MC),!,
	create_more_compounds(Compounds,[MC|Done],More).



compounds_in_compartment([],_CP,_CPN,Found,Found).


compounds_in_compartment([C|Compounds],Compartment,
			 CompartmentName,Done,Found) :-
	important_compound(C,_Name,_Class),
	member(C,Compartment),!,
	compounds_in_compartment(Compounds,Compartment,
				 CompartmentName,Done,Found).


compounds_in_compartment([C|Compounds],Compartment,
			 CompartmentName,Done,Found) :-
	important_compound(C,Name,_Class),
	NF = not_found(C,Name),!,
	compounds_in_compartment(Compounds,Compartment,
				 CompartmentName,[NF|Done],Found).


check_reactions([],_Used,Found,Found).

check_reactions([R1|Rs],Used,Done,Found) :-
	member(R1,Used),
	multiple_atom_concat([R1,' used'],RString),!,
	check_reactions(Rs,Used,[RString|Done],Found).

check_reactions([R1|Rs],Used,Done,Found) :-
	multiple_atom_concat([R1,' not used'],RString),!,
	check_reactions(Rs,Used,[RString|Done],Found).



pathway_checker(Knockout,Added,Allowed) :-
	run_yeast_reactions(Knockout,Added,Allowed,NotReached,_Ran),
	writelist(NotReached).
	%metabolic_pathway_display(Pathway,Cytosol).

check_non_experimentals([],Broken,Broken).


check_non_experimentals([NE|NEs],Done,AllBroken) :-
	NE = non_experimental(Orf,_Reason),
	write('processing':NE),nl,nl,
	run_yeast_reactions(Orf,[],[],NotReached,_Ran),!,
	update_broken_list(NE,NotReached,Done,NewDone),!,
	check_non_experimentals(NEs,NewDone,AllBroken).


update_broken_list(NE,[],Done,Done) :-
	write(not:NE),nl.

update_broken_list(NE,NotReached,Done,[Broken|Done]) :-
	NE = non_experimental(Orf,Reason),
	Broken = broken(Orf,Reason,NotReached).

all_orf_model_results([],Results,Results).


all_orf_model_results([Orf|Orfs],Done,Results) :-
	run_yeast_reactions(Orf,[],[],NotReached,_Ran),
	determine_growth(NotReached,Growth),
	write(Orf:Growth),nl,
	MR = model_result(Orf,Growth,NotReached),!,
	all_orf_model_results(Orfs,[MR|Done],Results).


all_model_growths(Growths) :-
	all_model_orfs(Orfs),
	all_orf_model_results(Orfs,[],Growths),
	open('full_model_results.pl',write,Stream),
	write_file_terms(Stream,Growths),
	close(Stream).

all_break_model_growths(Growths) :-
	all_break_orfs(Orfs),
	all_orf_model_results(Orfs,[],Growths),
	writelist(Growths).

all_no_effects :-
	findall(model_result(_Orf,'growth',_R),
		model_result(_Orf,'growth',_R),Growths),
	open('full_model_results.pl',write,Stream),
	write_file_terms(Stream,Growths),
	close(Stream).
	

all_model_orfs(Orfs) :-
	findall(O,orf_fact(O,_E,_EC,_G,_GD,_R),Os),
	make_set(Os,[],Orfs).

determine_growth([],growth).

determine_growth(_NotFound,'no growth').



all_brokens(Brokens) :-
	all_non_experimentals(NE),
	check_non_experimentals(NE,[],Brokens),
	open('confirmation_set.pl',write,Stream),
	write_file_terms(Stream,Brokens),
	close(Stream).


reaction_side_display([R]) :-
	R = reactant(Loc,_S,Compound),
	compound_name(Compound,Name),
	write(Compound:Name),nl,
	write(Loc).

reaction_side_display([R]) :-
	R = reactant(_S,Compound),
	compound_name(Compound,Name),
	write(Compound:Name).



reaction_side_display([R|Reactants]) :-
	R = reactant(Loc,_S,Compound),
	compound_name(Compound,Name),
	write(Compound:Name),nl,
	write(Loc),nl,
	write('+'),nl,
	reaction_side_display(Reactants).

reaction_side_display([R|Reactants]) :-
	R = reactant(_S,Compound),
	compound_name(Compound,Name),
	write(Compound:Name),nl,
	write('+'),nl,
	reaction_side_display(Reactants).


compound_name(Compound,Name) :-
	l_compound_name(Compound,[Name|_]).

compound_name(Compound,Name) :-
	yeast_model_metabolite(Compound,Name).

compound_name(Compound,Compound).


write_reaction(Reaction) :-
	reaction(Reaction,LHS,OP,RHS),
	write('reaction: '),write(Reaction),nl,nl,
	reaction_side_display(LHS),nl,
	write(OP),nl,
	reaction_side_display(RHS),nl,
	write('---------------------------------------------------------------------'),nl,nl.


write_reaction(Reaction) :-
	reaction(sce,Reaction,LHS,OP,RHS),
	write('kegg sce reaction: '),write(Reaction),nl,nl,
	reaction_side_display(LHS),nl,
	write(OP),nl,
	reaction_side_display(RHS),nl,
	write('---------------------------------------------------------------------'),nl,nl.

write_reaction(Reaction) :-
	reaction(kegg,Reaction,LHS,OP,RHS),
	write('kegg reaction: '),write(Reaction),nl,nl,
	reaction_side_display(LHS),nl,
	write(OP),nl,
	reaction_side_display(RHS),nl,
	write('---------------------------------------------------------------------'),nl,nl.


reactions_from_locations([],R,R).

reactions_from_locations([L|Locations],Done,Rs) :-
	L = location(Reaction,_S),!,
	reactions_from_locations(Locations,[Reaction|Done],Rs).


all_reactions_from_locations(Compound,Reactions) :-
	locations(Compound,Locations),
	reactions_from_locations(Locations,[],Reactions).

write_reactions([]).

write_reactions([R|Reactions]) :-
	write_reaction(R),!,
	write_reactions(Reactions).


compound_reactions(Compound) :-
	findall(R,(reaction(R,LHS,_OP,RHS),
		   reactants_from_all_lists([LHS,RHS],[],Rs),
		   member(Compound,Rs)),
		Reactions),
		make_set(Reactions,[],CReactions),
	write_reactions(CReactions).


compound_connectivity(Compound,Connectivity) :-
	findall(R,(reaction(R,LHS,_OP,RHS),
		   reactants_from_all_lists([LHS,RHS],[],Rs),
		   member(Compound,Rs)),
		Reactions),
		make_set(Reactions,[],CReactions),
		length(CReactions,Connectivity).


unusual_orf(Orf) :-
	var(Orf),
	orf_fact(O,E,EC,G,GD,R),
	name(O,[S|_]),
	name(Char,[S]),Char \== 'Y',
	Orf = orf_fact(O,E,EC,G,GD,R).

unusual_orf(Orf) :-
	Orf == orf_fact(O,_E,_EC,_G,_GD,_R),
	name(O,[S|_]),
	name(Char,[S]),Char \== 'Y'.



unknown_orf(Orf) :-
	var(Orf),
	orf_fact(O,E,EC,G,GD,R),
	name(O,[S|_]),
	name(Char,[S]),Char == 'U',
	Orf = orf_fact(O,E,EC,G,GD,R).

unknown_orf(Orf) :-
	Orf == orf_fact(O,_E,_EC,_G,_GD,_R),
	name(O,[S|_]),
	name(Char,[S]),Char == 'U'.




unknown_orf(UNum,E,EC,G,GD,R) :-
	var(UNum),
	orf_fact(Orf,E,EC,G,GD,R),
	name(Orf,[S|ARest]),
	name(Char,[S]),
	Char == 'U',
	conc(ANum,[_Last],ARest),
	name(UNum,ANum).

unknown_orf(Orf,E,EC,G,GD,R) :-
	nonvar(Orf),
	orf_fact(Orf,E,EC,G,GD,R),
	name(Orf,[S|_ARest]),
	name(Char,[S]),
	Char == 'U'.


all_unusual_orfs(AllOrfs) :-
	findall(Orf,unusual_orf(Orf),AllOrfs).

all_unknown_orfs(AllOrfs) :-
	findall(Orf,unknown_orf(Orf),AllOrfs).



display_unusual_orfs :-
	all_unusual_orfs(AllOrfs),
	display_unusual_orf_info(AllOrfs).

display_unknown_orfs :-
	all_unknown_orfs(AllOrfs),
	display_unusual_orf_info(AllOrfs).



display_unusual_orf_info([]).

display_unusual_orf_info([Orf|Orfs]) :-
	Orf = orf_fact(O,E,_EC,_G,GD,R),
	R \== none,
	write(orf:O),nl,
	write(enzyme:E),nl,
	write('gene description':GD),nl,
	write_reaction(R),!,
	display_unusual_orf_info(Orfs).

display_unusual_orf_info([Orf|Orfs]) :-
	Orf = orf_fact(O,E,_EC,_G,GD,_R),
	write(orf:O),nl,
	write(enzyme:E),nl,
	write('gene description':GD),nl,
	write('reaction:none'),nl,
	write('--------------------------------------'),nl,!,
	display_unusual_orf_info(Orfs).


known_orf_unknown_enzyme(Orf) :-
	orf_fact(O,none,EC,G,GD,R),
	Orf = orf_fact(O,none,EC,G,GD,R),trace,
	\+unknown_orf(Orf).


all_unknown_orfs([]).

all_unknown_orfs([Orf|Orfs]) :-
	unknown_orf(Orf),!,
	all_unknown_orfs(Orfs).


all_orf_facts_reaction(R,Orfs) :-
	findall(orf_fact(Orf,E,EC,G,GD,R),orf_fact(Orf,E,EC,G,GD,R),Orfs).


all_reactions_in_unknowns(Reactions) :-
	findall(Orf,known_orf_unknown_enzyme(Orf),Orfs),
	make_set(Orfs,[],Reactions).


no_reaction(Orf,E,GD) :-
	orf_fact(Orf,E,_EC,_G,GD,none).


all_unknown_orf_facts([]).

all_unknown_orf_facts([O|OrfFacts]) :-
	O = orf_fact(Orf,E,EC,G,GD,R),
	unknown_orf(Orf,E,EC,G,GD,R),
	name(Orf,[_U|AOrf]),
	conc(ANum,[_Last],AOrf),
	name(UNum,ANum),
	UNum < 109,!,
	all_unknown_orf_facts(OrfFacts).



genes_from_organisms(Organism,Enzyme,OtherGenes) :-
	all_enzyme_genes(Enzyme,Genes),
	remove(l_enzyme_genes(Enzyme,Organism,_YGenes),Genes,GeneStrings),
	all_gene_names(GeneStrings,[],OtherGenes).


all_enzyme_genes(Enzyme,Genes) :-
	findall(l_enzyme_genes(Enzyme,Organism,Gene),
		l_enzyme_genes(Enzyme,Organism,Gene),Genes).


gene_names(_E,_O,[],GeneNames,GeneNames).

gene_names(Enzyme,Organism,[G|GeneStrings],Done,GeneNames) :-
	create_gene_structure(G,Orf,Alt),
	GName = gene_name(Organism,Enzyme,Orf,Alt),!,
	gene_names(Enzyme,Organism,GeneStrings,[GName|Done],GeneNames).
	

all_gene_names([],GeneNames,GeneNames).

all_gene_names([LE|LEs],Done,GeneNames) :-
	LE = l_enzyme_genes(Enzyme,Organism,GeneStrings),
	gene_names(Enzyme,Organism,GeneStrings,[],GNames),
	conc(Done,GNames,NewDone),!,
	all_gene_names(LEs,NewDone,GeneNames).


orfs_only_from_kegg(Organism,KeggOrfs) :-
	findall(kegg_orf(O1,E1,G1),
        (orf_enzyme(Organism,O1,G1,E1),
                    \+orf_fact(O1,E1,_EC2,_G2,_GD2,_R)),Es),
        make_set(Es,[],KeggOrfs).


orfs_no_reactions(Organism,KeggOrfs) :-
	findall(kegg_orf(O1,E1,G1),
        (orf_enzyme(Organism,O1,G1,E1),
                    orf_fact(O1,E1,_EC2,_G2,_GD2,none)),Es),
        make_set(Es,[],KeggOrfs).




kegg_reactions_from_enzyme(Enzyme,ReactionSet) :-
	findall(RNum,(l_reaction_enzyme(RNum,E),
                      member(Enzyme,E)),Reactions),
	make_set(Reactions,[],ReactionSet).


organism_reactions(_O,[],Reactions,Reactions).

organism_reactions(Organism,[RNum|RNums],Done,Reactions) :-
	reaction(Organism,RNum,_LHS,_OP,_RHS),!,
	organism_reactions(Organism,RNums,[RNum|Done],Reactions).

organism_reactions(Organism,[_RNum|RNums],Done,Reactions) :-
	!,
	organism_reactions(Organism,RNums,Done,Reactions).


all_new_reactions(_O,[],Rs,Rs).

all_new_reactions(Organism,[KO|KOs],Done,NewRs) :-
	KO = kegg_orf(_O1,E1,_G1),
        kegg_reactions_from_enzyme(E1,Rs),
	organism_reactions(Organism,Rs,[],ReducedRs),
	conc(ReducedRs,Done,NewDone),!,
	all_new_reactions(Organism,KOs,NewDone,NewRs).


find_kegg_orf_reactions(Organism,NewFacts,ReNumbered) :-
	orfs_only_from_kegg(Organism,KeggOrfs),
	orfs_no_reactions(Organism,NORs),
	conc(KeggOrfs,NORs,AllKeggOrfs),
	%write(AllKeggOrfs),get0(10),
	all_new_reactions(Organism,AllKeggOrfs,[],KReactions),
	%write(KReactions),get0(10),
	make_set(KReactions,[],KReactionSet),
	reorganise_reactions(841,KReactionSet,[],Mappings,[],ReNumbered),
	writelist(ReNumbered),get0(10),	
	all_kegg_orf_facts(Organism,AllKeggOrfs,Mappings,[],NewFacts),
	write(NewFacts),get0(10).


add_kegg_reactions(Organism) :-
	find_kegg_orf_reactions(Organism,NewFacts,NewReactions),
	open('added_kegg_facts.pl',write,Stream1),
	write_file_terms(Stream1,NewFacts),
	close(Stream1),
	open('added_kegg_reactions.pl',write,Stream2),
	write_file_terms(Stream2,NewReactions),
	close(Stream2).



reorganise_reactions(_I,[],Mapping,Mapping,ReNumbered,ReNumbered).

reorganise_reactions(I,[RNum|RNums],MDone,Mapping,RDone,ReNumbered) :-
	reaction(_O,RNum,LHS,OP,RHS),
	reorganise_side(LHS,[],OLHS),
	reorganise_side(RHS,[],ORHS),
	NewMap = mapping(I,RNum),
	ReNumR = reaction(I,OLHS,OP,ORHS),
	I2 is I + 1,!,
	reorganise_reactions(I2,RNums,[NewMap|MDone],Mapping,[ReNumR|RDone],ReNumbered).


reorganise_side([],R,R).

reorganise_side([R|Reactants],Done,Rs) :-
	R = reactant(N,C),
	NewR = reactant(cytosol,N,C),!,
	reorganise_side(Reactants,[NewR|Done],Rs).


create_orf_facts_from_kegg_reactions([],_Mappings,_Orf,_Enzyme,_Gene,
				     NewFacts,NewFacts).

create_orf_facts_from_kegg_reactions([RNum|RNums],Mappings,Orf,Enzyme,Gene,
				     Done,NewFacts) :-
	find_mapping(RNum,Mappings,MapNum),
	make_enzyme_class(Enzyme,EClass),
	NewOrfFact = orf_fact(Orf,Enzyme,EClass,Gene,kegg,MapNum),!,
	create_orf_facts_from_kegg_reactions(RNums,Mappings,Orf,Enzyme,Gene,
				     [NewOrfFact|Done],NewFacts).

create_orf_facts_from_kegg_reactions([_RNum|RNums],Mappings,Orf,Enzyme,Gene,
				     Done,NewFacts) :-
        !,
	create_orf_facts_from_kegg_reactions(RNums,Mappings,Orf,Enzyme,Gene,
				     Done,NewFacts).


all_kegg_orf_facts(_O,[],_M,NewFacts,NewFacts).

all_kegg_orf_facts(Organism,[KO|KOs],Mappings,Done,NewFacts) :-
	KO = kegg_orf(O,E,G),
        kegg_reactions_from_enzyme(E,Rs),
	organism_reactions(Organism,Rs,[],ReducedRs),
	create_orf_facts_from_kegg_reactions(ReducedRs,Mappings,O,E,G,[],Facts),
	conc(Facts,Done,NewDone),!,
	all_kegg_orf_facts(Organism,KOs,Mappings,NewDone,NewFacts).


find_mapping(RNum,[M|_Mappings],MappedNum) :-
	M = mapping(MappedNum,RNum).

find_mapping(RNum,[_M|Mappings],MappedNum) :-
	!,find_mapping(RNum,Mappings,MappedNum).



ran_reactions([],_RanReactions,Ran,Ran).

ran_reactions([R|Reactions],RanReactions,Done,Ran) :-
	member(R,RanReactions),!,
	ran_reactions(Reactions,RanReactions,[R|Done],Ran).

ran_reactions([_R|Reactions],RanReactions,Done,Ran) :-
	!,ran_reactions(Reactions,RanReactions,Done,Ran).


all_non_experimentals(Orfs) :-
	clauses_from_file('no_experimental_evidence.pl',Orfs).


generic_map_number(MapNum) :-
	name(MapNum,[M,A,P|_Rest]),
	name('MAP',[M,A,P]).


all_map_numbers(Compound,MapNumbers) :-
	findall(MapNum,
		l_compound_pathway(Compound,path(MapNum,_Name)),
		MapNumbers).
				   

contains_generic_map_num([Mnum|_Mnums]) :-
	generic_map_number(Mnum).

contains_generic_map_num([_Mnum|Mnums]) :-
	!,contains_generic_map_num(Mnums).


all_break_orfs(Orfs) :-
	findall(O,model_result(O,'no growth',_R),Orfs).


establish_inner_recovery_point(I,Met,Metabolites) :-
	Res is I mod 10,
	Res == 1,
	exp_generator_directory(ExpDir),
	multiple_atom_concat([ExpDir,'inner_recovery_point.pl'],IRPFile),
	IRP = inner_recovery_point(Met,Metabolites),
	open(IRPFile,write,Stream),
	write_file_terms(Stream,[IRP]),
	close(Stream).

establish_inner_recovery_point(_I,_Met,_Metabolites).

%% alternative for wet evidence genes

establish_wet_inner_recovery_point(I,Met,Metabolites) :-
	Res is I mod 10,
	Res == 1,
	exp_generator_directory(ExpDir),
	multiple_atom_concat([ExpDir,'wet_inner_recovery_point.pl'],IRPFile),
	IRP = inner_recovery_point(Met,Metabolites),
	open(IRPFile,write,Stream),
	write_file_terms(Stream,[IRP]),
	close(Stream).

establish_wet_inner_recovery_point(_I,_Met,_Metabolites).


find_recovered_growths(_I,_O,_Gene,[]).

find_recovered_growths(I,OutFile,Gene,[Met|Metabolites]) :-
	establish_inner_recovery_point(I,Met,Metabolites),
	ensure_list(Met,ListMet),%trace,%write('about to run model'),nl,nl,
	run_yeast_reactions(Gene,ListMet,[],NotReached,_Ran),%write('determining growth'),nl,nl,
	determine_growth(NotReached,Growth),
	write(Gene:Met:Growth),nl,
	(   Growth == 'growth',
	    open(OutFile,append,Stream),
	    write_file_terms(Stream,[recovered_growth(Gene,Met)]),nl(Stream),
	    close(Stream),
	    !,
	    I2 is I + 1,
	    find_recovered_growths(I2,OutFile,Gene,Metabolites)

	;   

	    I2 is I + 1,!,
	    find_recovered_growths(I2,OutFile,Gene,Metabolites)).
	


%% alternative for wet evidence genes

find_wet_recovered_growths(_I,_O,_Gene,[]).

find_wet_recovered_growths(I,OutFile,Gene,[Met|Metabolites]) :-
	establish_wet_inner_recovery_point(I,Met,Metabolites),
	ensure_list(Met,ListMet),%trace,%write('about to run model'),nl,nl,
	run_yeast_reactions(Gene,ListMet,[],NotReached,_Ran),%write('determining growth'),nl,nl,
	determine_growth(NotReached,Growth),
	write(Gene:Met:Growth),nl,
	(   Growth == 'growth',
	    open(OutFile,append,Stream),
	    write_file_terms(Stream,[recovered_growth(Gene,Met)]),nl(Stream),
	    close(Stream),
	    !,
	    I2 is I + 1,
	    find_wet_recovered_growths(I2,OutFile,Gene,Metabolites)

	;   

	    I2 is I + 1,!,
	    find_wet_recovered_growths(I2,OutFile,Gene,Metabolites)).

ensure_list(List1,List2) :-
	list(List1),
	List2 = List1.

ensure_list(List1,[List1]).


conjectures_and_experiments(NumberFind,Organism,UnknownGene,Type,Method,CJFile,ExpFile) :-
	all_recovered_growths(UnknownGene,OutFile),
	write('passed all recovered growths'),nl,nl,
	all_orf_info(UnknownGene,OrfInfo),
	find_knocked_out_reactions_and_enzymes(UnknownGene,_Enzymes,Reactions),
	knocked_out_orf_info(OrfInfo,Reactions,[],KOOrfInfo),
	clauses_from_file(OutFile,Recovered),
	load_full_metabolite_graph,
	all_possible_orfs_and_conjectures(NumberFind,Organism,KOOrfInfo,[],
						   Type,Method,[],Conjectures),
	write('passed possible orfs'),nl,nl,%trace,
	construct_experiments(Conjectures,Recovered,[],AllExperiments),%trace,
	real_conjectures(Conjectures,[],AllRealConjectures),
	make_set(AllExperiments,[],Experiments),
	make_set(AllRealConjectures,[],RealConjectures),
	open(CJFile,append,Stream1),
	open(ExpFile,append,Stream2),
	write_file_terms(Stream1,RealConjectures),
	write_file_terms(Stream2,Experiments),
	close(Stream1),
	close(Stream2).



all_orf_info(Gene,OrfInfo) :-
	findall(orf_info(Gene,Enzyme,Reaction),
		orf_fact(Gene,Enzyme,_EC,_G,_GD,Reaction),OrfInfo).



knocked_out_orf_info([],_R,KO,KO).

knocked_out_orf_info([OI|OrfInfo],Reactions,Done,AllKO) :-
	OI = orf_info(_O,_E,R),
	member(R,Reactions),!,
	knocked_out_orf_info(OrfInfo,Reactions,[OI|Done],AllKO).




%% alternative for wet evidence genes

wet_conjectures_and_experiments(KG,CJFile,ExpFile) :-
	all_wet_recovered_growths(KG,OutFile),
	write('passed all recovered growths':KG),nl,nl,
	clauses_from_file(OutFile,Recovered),
	write('passed loading recovered growth'),nl,nl,%trace,
	all_orf_info(KG,OrfInfo),
	find_knocked_out_reactions_and_enzymes(KG,_Enzymes,Reactions),
	knocked_out_orf_info(OrfInfo,Reactions,[],KOOrfInfo),	
	make_wet_conjectures(KOOrfInfo,[],Conjectures),
	write('made wet conjectures':KG),nl,nl,
	construct_wet_experiments(Conjectures,Recovered,[],Experiments),%trace,
	write('constructed wet experiments':KG),nl,nl,
	open(CJFile,append,Stream1),
	write_file_terms(Stream1,Conjectures),
	close(Stream1),
	open(ExpFile,append,Stream2),
	write_file_terms(Stream2,Experiments),
	close(Stream2),
	write('written to wet experiment and conjecture files':KG),nl,nl.


real_conjectures([],RealCons,RealCons).

real_conjectures([C|Conjectures],Done,RealCons) :-
	C = conjecture(OldGene,NewGene,Enzyme,Reaction,_Rank),
	RC = conjecture(OldGene,NewGene,Enzyme,Reaction),!,
	real_conjectures(Conjectures,[RC|Done],RealCons).
	

one_conjecture(_C,[],RDone,Experiments) :-
	reverse(RDone,Experiments).

one_conjecture(Conjecture,[R|Recovered],Done,Experiments) :-
	R = recovered_growth(_G,Mets),
	%wild_type_reactions(Ran),
	%construct_metabolite_graph_from_reactions(Ran),
	Conjecture = conjecture(OldGene,NewGene,_Enzyme,_Reaction,Rank),%trace,
	find_knocked_out_reactions_and_enzymes(OldGene,Reactions,_Enzymes),
	find_reaction_distances_and_paths(1,Mets,Reactions,Distance,Paths),
	reactions_from_paths(Paths,[],ReactionsAndPaths),%trace,
	Exp = experiment(phenotypic_effect(NewGene,Mets),Rank,Distance,ReactionsAndPaths),!,
	one_conjecture(Conjecture,Recovered,[Exp|Done],Experiments).

% one wet conjecture

one_wet_conjecture(_C,[],RDone,Experiments) :-
	reverse(RDone,Experiments).

one_wet_conjecture(Conjecture,[R|Recovered],Done,Experiments) :-
	R = recovered_growth(_G,Mets),
	%wild_type_reactions(Ran),
	%construct_metabolite_graph_from_reactions(Ran),
	Conjecture = conjecture(Gene,_Enzyme,_Reaction),%trace,
	find_knocked_out_reactions_and_enzymes(Gene,Reactions,_Enzymes),
	find_reaction_distances_and_paths(1,Mets,Reactions,Distance,Paths),
	reactions_from_paths(Paths,[],ReactionsAndPaths),%trace,
	Exp = experiment(phenotypic_effect(Gene,Mets),Distance,ReactionsAndPaths),!,
	one_wet_conjecture(Conjecture,Recovered,[Exp|Done],Experiments).



reactions_from_paths([],Reactions,Reactions).

reactions_from_paths([P|Paths],Done,Reactions) :-
	get_reactions_from_path(P,PRs),
	PR = reaction_details(P,PRs),!,
	reactions_from_paths(Paths,[PR|Done],Reactions).



find_reaction_distances_and_paths(Threshold,Mets,Reactions,ShortestD,Paths) :-
	%make_nodes(Mets,[],Nodes),
	all_reaction_distances(Threshold,Mets,Reactions,[],RDists),
	quicksort_relation(3,descending,RDists,Sorted),
	Sorted = [F|_Rest],
	F = shortest_distance(_N,_R,ShortestD,_Path),
	collect_all_paths(RDists,[],Paths).


collect_all_paths([],Paths,Paths).

collect_all_paths([RD|RDists],Done,Paths) :-
	RD = shortest_distance(_N,R,_D,Path),
	RP = reaction_path(R,Path),!,
	collect_all_paths(RDists,[RP|Done],Paths).


find_reaction_distances_and_paths(_Threshold,_Mets,_Reaction,'no path','no path').


all_reaction_distances(_T,_Mets,[],RDists,RDists).

all_reaction_distances(Threshold,Mets,[R|Reactions],Done,RDists) :-
	reaction_metabolite_distances(Threshold,Mets,R,[],RDs),
	conc(Done,RDs,NewDone),!,
	all_reaction_distances(Threshold,Mets,Reactions,NewDone,RDists).


reaction_metabolite_distances(_Threshold,[],_Reaction,RDist,RDist).

reaction_metabolite_distances(Threshold,[Met|Mets],Reaction,Done,RDists) :-
	find_shortest_distance(Threshold,148,Met,Reaction,ShortestD,Path),
	RD = shortest_distance(Met,Reaction,ShortestD,Path),!,
	reaction_metabolite_distances(Threshold,Mets,Reaction,[RD|Done],RDists).

reaction_metabolite_distances(Threshold,[_Met|Mets],Reaction,Done,RDists) :-
	!,
	reaction_metabolite_distances(Threshold,Mets,Reaction,Done,RDists).


make_nodes([],Nodes,Nodes).

make_nodes([Met|Mets],Done,Nodes) :-
	construct_nodes(Met,MNodes),
	conc(MNodes,Done,NewDone),!,
	make_nodes(Mets,NewDone,Nodes).


construct_nodes(Met,[node(cytosol,Met),node(mitochondrion,Met)]) :-
	node(cytosol,Met),
	node(mitochondrion,Met).

construct_nodes(Met,[node(cytosol,Met)]) :-
	node(cytosol,Met).

construct_nodes(Met,[node(mitochondrion,Met)]) :-
	node(mitochondrion,Met).

construct_nodes(_Met,[]).


find_shortest_distance(Threshold,_Max,Met,Reaction,ShortestD,Path) :-
	reaction_distance(Threshold,Met,Reaction,ShortestD,Path),
	write(Threshold:foundpath),nl.

find_shortest_distance(Threshold,Max,_Met,_Reaction,_ShortestD,_Path) :-
	Threshold > Max,write(Threshold:stopping),
	!,fail.

find_shortest_distance(Threshold,Max,Met,Reaction,ShortestD,Path) :-
	write(Threshold:increasing_threshold),nl,
	Thresh2 is Threshold + 1,!,
	find_shortest_distance(Thresh2,Max,Met,Reaction,ShortestD,Path).

	


construct_experiments([],_R,Experiments,Experiments).

construct_experiments([C|Conjectures],Recovered,Done,Experiments) :-
	one_conjecture(C,Recovered,[],CExpts),
	conc(Done,CExpts,NewDone),!,
	construct_experiments(Conjectures,Recovered,NewDone,Experiments).
	

% constructing wet experiments

construct_wet_experiments([],_R,Experiments,Experiments).

construct_wet_experiments([C|Conjectures],Recovered,Done,Experiments) :-
	one_wet_conjecture(C,Recovered,[],CExpts),
	conc(Done,CExpts,NewDone),!,
	construct_wet_experiments(Conjectures,Recovered,NewDone,Experiments).



all_recovered_growths(UnknownGene,OutFile) :-
	exp_generator_directory(ExpDir),
	multiple_atom_concat([ExpDir,'single_compound_lists.pl'],SCLFile),
	clauses_from_file(SCLFile,CompoundList),
	CompoundList = [single_compound_lists(Compounds)],
	multiple_atom_concat([ExpDir,UnknownGene,'_recovered_growth.pl'],OutFile),
	findall_recovered_growths(UnknownGene,Compounds,Compounds,OutFile).


%% alternative for wet evidence genes

all_wet_recovered_growths(Gene,OutFile) :-
	exp_generator_directory(ExpDir),
	multiple_atom_concat([ExpDir,'single_compound_lists.pl'],SCLFile),
	clauses_from_file(SCLFile,CompoundList),
	CompoundList = [single_compound_lists(Compounds)],
	PLM = previous_level_metabolites(Compounds),
	open('previous_level_metabolites.pl',write,Stream1),
	write_file_terms(Stream1,[PLM]),
	close(Stream1),
	multiple_atom_concat([ExpDir,Gene,'_recovered_growth.pl'],OutFile),
	findall_wet_recovered_growths(Gene,Compounds,OutFile).


%% alternative for wet evidence genes

findall_wet_recovered_growths(Gene,LevelMetabolites,OutFile) :-
	find_wet_recovered_growths(1,OutFile,Gene,LevelMetabolites),
        write('finished recovered growths'),nl,nl,
    (   write('testing for file'),nl,nl,
	    exists_file(OutFile)
	;   
	    write('making next level metabolites'),nl,nl,
	    exp_generator_directory(ExpDir),
	    multiple_atom_concat([ExpDir,'compounds_from_reactions.pl'],CRFile),
	    multiple_atom_concat([ExpDir,'previous_level_metabolites.pl'],PLMFile),
            clauses_from_file(CRFile,[CR]),
            clauses_from_file(PLMFile,[PLM]),
	    CR = compounds_from_reactions(Metabolites),
	    PLM = previous_level_metabolites(PreviousMetabolites),
	    all_combinations(Metabolites,PreviousMetabolites,[],NextLevelMetabolites),
	    NLM = previous_level_metabolites(NextLevelMetabolites),
	    open(PLMFile,write,Stream1),
	    write_file_terms(Stream1,[NLM]),
	    close(Stream1),!,
	    findall_wet_recovered_growths(Gene,NextLevelMetabolites,OutFile)).


findall_recovered_growths(Gene,Metabolites,SingleCLists,OutFile) :-
	find_recovered_growths(1,OutFile,Gene,Metabolites),
	(   exists_file(OutFile)
	    
	;   
	    all_combinations(Metabolites,SingleCLists,[],NextLevelMetabolites),
	    findall_recovered_growths(Gene,NextLevelMetabolites,SingleCLists,OutFile)).




wet_gene_experiments(Gene) :-
	all_wet_recovered_growths(Gene,OutFile),
	clauses_from_file(OutFile,Recovered),
	construct_wet_experiments(Recovered,[],Experiments),
	multiple_atom_concat([Gene,'_experiments.pl'],ExpFile),
	open(ExpFile,write,Stream),
	write_file_terms(Stream,Experiments),
	close(Stream).
	
construct_wet_experiments([],RDone,Experiments) :-
	reverse(RDone,Experiments).

construct_wet_experiments([R|Recovered],Done,Experiments) :-
	R = recovered_growth(Gene,Mets),
	Exp = experiment(phenotypic_effect(Gene,Mets)),
	construct_wet_experiments(Recovered,[Exp|Done],Experiments).

all_gene_recovered_growths([]).

all_gene_recovered_growths([G|Genes]) :-
    all_recovered_growths(G,_F),!,
    all_gene_recovered_growths(Genes).


all_compounds_from_reactions(Compounds) :-
	findall(ReactantList,(reaction(_RNum,LHS,_OP,RHS),
			      reactants_from_all_lists([LHS,RHS],[],ReactantList)),
				       ReactantLists),
	full_compound_list(ReactantLists,[],Compounds).


all_reactant_lists_from_reactions([],RLs,RLs).

all_reactant_lists_from_reactions([R|Reactions],Done,RLs) :-
	reaction(R,LHS,_OP,RHS),
	reactants_from_all_lists([LHS,RHS],[],ReactantList),!,
	all_reactant_lists_from_reactions(Reactions,[ReactantList|Done],RLs).


all_compounds_from_reaction_set(Reactions,Compounds) :-
	all_reactant_lists_from_reactions(Reactions,[],ReactantLists),
	full_compound_list(ReactantLists,[],Compounds).


create_compound_list([],B,B).

create_compound_list([C|Compounds],Building,Built) :-
	\+member(C,Building),!,
	create_compound_list(Compounds,[C|Building],Built).

create_compound_list([_C|Compounds],Building,Built) :-
	!,create_compound_list(Compounds,Building,Built).


full_compound_list([],Full,Full).

full_compound_list([CList|CLists],Building,FullList) :-
	create_compound_list(CList,Building,NewBuilding),!,
	full_compound_list(CLists,NewBuilding,FullList).



interesting_unknown_genes(['U51_']).
%interesting_unknown_genes(['U96_','U94_','U93_','U92_','U89_','U51_']).

interesting_known_genes(['YDR531W','YDR354W','YPR060C','YMR108W',
	                 'YJR148W','YOR074C','YNR016C','YPL028W','YDL055C']).


interesting_genes(Genes) :-
	interesting_unknown_genes(UG),
	interesting_known_genes(KG),
	conc(KG,UG,Genes).


interesting_recovered_growth :-
    interesting_genes(Genes),
    all_gene_recovered_growths(Genes).



all_enzymes_and_reactions(Orf,EnzymeReactions) :-
	findall(enzyme_reactions(E,R),orf_fact(Orf,E,_EC,_G,_GD,R),
		EnzymeReactions).


compile_reaction_enzyme_info([]).

compile_reaction_enzyme_info([ER|ERs]) :-
	ER = enzyme_reactions(Enz,Reaction),
	write('Enzyme Class: '),write(Enz),nl,nl,
	write_reaction(Reaction),nl,nl,!,
	compile_reaction_enzyme_info(ERs).
	


info_for_gene(Orf) :-
	write('Gene: '),write(Orf),nl,
	all_enzymes_and_reactions(Orf,ERs),
	compile_reaction_enzyme_info(ERs).


info_for_genes([]).

info_for_genes([O|Orfs]) :-
	info_for_gene(O),!,
	info_for_genes(Orfs).


all_recovered_growth(Growths) :-
	clauses_from_file('all_recovered_growth.pl',Growths).


add_name_to_growth([],Growths,Growths).

add_name_to_growth([G|Growths],Done,NewGrowths) :-
	G = recovered_growth(Gene,CID),
	l_compound_name(CID,[FName|_Rest]),
	NewGrowth = recovered_growth(Gene,CID,FName),!,
	add_name_to_growth(Growths,[NewGrowth|Done],NewGrowths).

add_name_to_growth([G|Growths],Done,NewGrowths) :-
	G = recovered_growth(Gene,CID),
	yeast_model_metabolite(CID,FName),
	NewGrowth = recovered_growth(Gene,CID,FName),!,
	add_name_to_growth(Growths,[NewGrowth|Done],NewGrowths).


names_for_growth(NewGrowths) :-
	all_recovered_growth(Growths),
	add_name_to_growth(Growths,[],NewGrowths).

	

reactant_locations([],Locations,Locations).

reactant_locations([R|Reactants],Done,Locations) :-
	R = reactant(Loc,_S,_C),!,
	reactant_locations(Reactants,[Loc|Done],Locations).


reaction_locations(RNum,Locations) :-
	reaction(RNum,LHS,_OP,RHS),
	reactant_locations(LHS,[],LLocs),
	reactant_locations(RHS,[],RLocs),
	conc(LLocs,RLocs,Locations).

mitochondrion_reaction(RNum) :-
	reaction_locations(RNum,Locations),
	\+member(external,Locations),
	\+member(cytosol,Locations).

cytosol_reaction(RNum) :-
	reaction_locations(RNum,Locations),
	\+member(external,Locations),
	\+member(mitochondrion,Locations).


reaction_match(R1,R2) :-
	reaction(R1,LHS1,_OP1,RHS1),
	reaction(R2,LHS2,_OP2,RHS2),
	R1 \== R2,
	reactants_from_all_lists([LHS1,RHS1],[],Rs1),
	reactants_from_all_lists([LHS2,RHS2],[],Rs2),
	equivalent_list(Rs1,Rs2).



all_growth_chemicals(Growths) :-
	findall(CID,recovered_growth(_K,CID,_Name),CIDs),
	make_set(CIDs,[],Growths).


growth_knockouts(CID,Knockouts) :-
	findall(K,recovered_growth(K,CID,_N),Knockouts).


all_growth_info(CID,INFO) :-
	growth_knockouts(CID,Knockouts),%trace,
	name_formula_and_cas_entry(CID,CNames,CFormula,CASEntry),%trace,
	INFO = growth_information(CID,Knockouts,CFormula,CNames,CASEntry).


name_formula_and_cas_entry(CID,Names,Formula,CASEntry) :-
	l_compound_name(CID,Names),
	l_compound_formula(CID,Formula),
	l_compound_dblinks(CID,cas,_,CASEntry).

name_formula_and_cas_entry(CID,Names,Formula,[]) :-
	l_compound_name(CID,Names),
	l_compound_formula(CID,Formula).

name_formula_and_cas_entry(CID,Names,'',[]) :-
	l_compound_name(CID,Names).

name_formula_and_cas_entry(CID,[Name],'',[]) :-
	yeast_model_metabolite(CID,Name).


growth_information([],I,I).

growth_information([C|CIDs],Done,Information) :-
	all_growth_info(C,Info),!,
	growth_information(CIDs,[Info|Done],Information).


collect_growth_info(Info) :-
	all_growth_chemicals(CIDs),
	growth_information(CIDs,[],Info),
	open('recovered_growth.txt',write,Stream),
	display_growth_info(Stream,Info),
	close(Stream).


display_growth_info(_S,[]).

display_growth_info(Stream,[I|Info]) :-
	I = growth_information(CID,Knockouts,Formula,Names,CASEntry),
	write(Stream,'KeggID: '),write(Stream,CID),nl(Stream),
	write(Stream,'Knockouts:'),nl(Stream),
	write_file_list(Stream,Knockouts),nl(Stream),
	write(Stream,'Formula: '),write(Stream,Formula),nl(Stream),
	write(Stream,'CAS Entries:'),nl(Stream),
	write_file_list(Stream,CASEntry),nl(Stream),
	write(Stream,'Names: '),nl(Stream),
	write_file_list(Stream,Names),nl(Stream),nl(Stream),!,
	display_growth_info(Stream,Info).



orf_names_from_kegg(Orfs) :-
	clauses_from_file('added_kegg_facts.pl',Clauses),
	extract_gene_names(Clauses,[],GNames),
	make_set(GNames,[],Orfs).

orf_names_from_danish(Orfs) :-
	clauses_from_file('flat_orf_facts.pl',Clauses),
	extract_gene_names(Clauses,[],GNames),
	make_set(GNames,[],Orfs).


all_wet_biology([],Ev,Ev).

all_wet_biology([O|Orfs],Done,Ev) :-
	wet_biology_gene(O),!,
        all_wet_biology(Orfs,[O|Done],Ev).

all_wet_biology([_O|Orfs],Done,Ev) :-
        !,all_wet_biology(Orfs,Done,Ev).


extract_gene_names([],N,N).

extract_gene_names([O|Orfs],Done,Names) :-
	O = orf_fact(Orf,_E,_EC,_G,_GD,_R),!,
	extract_gene_names(Orfs,[Orf|Done],Names).



orf_viability(Orf,[lethal]) :-
	lethal(Orf).

orf_viability(Orf,[viable]) :-
	viable(Orf).

orf_viability(Orf,['lethal or viable']) :-
	lethal_viable(Orf).

orf_viability(Orf,[lethal,viable]) :-
	lethal(Orf),
	viable(Orf).

orf_viability(Orf,['lethal or viable',lethal]) :-
	lethal_viable(Orf),
	lethal(Orf).

orf_viability(Orf,['lethal or viable',viable]) :-
	lethal_viable(Orf),
	viable(Orf).

orf_viability(Orf,['lethal or viable',viable,lethal]) :-
	lethal_viable(Orf),
	viable(Orf),
	lethal(Orf).


orf_viability(Orf,[unchecked]) :-
	\+lethal(Orf),
	\+viable(Orf),
	\+lethal_viable(Orf).



all_orf_stats(ModelGrowth,Viability,OrfList) :-
	findall(orf_stats(Orf,ModelGrowth,Viability),
		(model_result(Orf,ModelGrowth,_NF),
		 orf_viability(Orf,Viability)),OrfList).


totaliser(Number) :-
	all_orf_stats(growth,[viable],GVList),
	all_orf_stats(growth,[lethal],GLList),
	all_orf_stats(growth,['lethal or viable'],GVorLList),
	all_orf_stats('no growth',[viable],NGVList),
	all_orf_stats('no growth',[lethal],NGLList),
	all_orf_stats('no growth',['lethal or viable'],NGVorLList),
	length(GVList,LN1),
	length(GLList,LN2),
	length(GVorLList,LN3),
	length(NGVList,LN4),
	length(NGLList,LN5),
	length(NGVorLList,LN6),
	Number is (LN1 + LN2 + LN3 + LN4 + LN5 + LN6).


all_growth_results(Gs) :-
	findall(model_result(O,G,NF),model_result(O,G,NF),Gs).


orf_viabilities(Orf,Viabilities) :-
	findall(Viability,orf_viability(Orf,Viability),Viabilities).


display_orf_viabilities([]).

display_orf_viabilities([M|ModelResults]) :-
	M = model_result(Orf,Growth,_NF),
	orf_viabilities(Orf,Viabilities),
	write(Orf:' '),write(Growth),write(' '),
	write(Viabilities),get0(10),
	display_orf_viabilities(ModelResults).


display_orf_model_predictions_and_viabilities :-
	all_growth_results(ModelResults),
	display_orf_viabilities(ModelResults).

model_stuff :-
	display_orf_model_predictions_and_viabilities.




all_orf_viabilities([],_V,OV,OV).

all_orf_viabilities([Orf|Orfs],Viability,Done,OV) :-
	orf_viability(Orf,Viability),
	OrfV = orf_viability(Orf,Viability),!,
	all_orf_viabilities(Orfs,Viability,[OrfV|Done],OV).

all_orf_viabilities([_Orf|Orfs],Viability,Done,OV) :-
	!,all_orf_viabilities(Orfs,Viability,Done,OV).



findall_orf_viabilities(Viability,OrfViabilities) :-
	all_orfs(Orfs),
	all_orf_viabilities(Orfs,Viability,[],OrfViabilities).
